
#include "ros/ros.h"
#include "igk_g416/data.h"
#include "igk_g416/netdata.h"
#include "serial/serial.h"
#include <std_msgs/String.h>
#include <iostream>
#include <vector>

std::vector<uint8_t> buff;
serial::Serial ser;
uint8_t SendData[8]={0};
uint8_t rev_ok = 1;
std::string port;
double RequestCycle;
int Baudrate,ID,Mode;
igk_g416::data Data;
igk_g416::netdata NetData;
/*************************************************
crc16校验码计算函数,计算算法：
1、设置crc校验寄存器的初值为0xffff;
2、将报文的第一个字节与crc寄存器的低字节异或，结果存入crc寄存器
3、判断crc的第一位是否为1，如果是1,crc右移1位后和0xa001异或，如果为0，则再移1位；
4、重复步骤3，直到完成8个位；
5、重复步骤2、3、4直到完成所有字节
6、返回计算结果
***********************************************/
uint16_t CalcCrc(unsigned char *chData, unsigned short uNo)
{
    uint16_t crc = 0xffff;
    uint16_t i, j;
    for(i = 0; i < uNo; i++)
    {
        crc ^= chData[i];
        for(j = 0; j < 8; j++)
        {
            if(crc & 1)
            {
                crc >>= 1;
                crc ^= 0xA001;
            }
            else
            {
                crc >>= 1;
            }
        }
    }
    return crc;
}

//检查Modbus CRC校验
//buf:完整数据包，包含crc的两个字节
//length：完整数据包，包含crc的两个字节
uint8_t CheckModbusCRC(uint8_t *buf,uint16_t length)
{
	uint16_t _crc;
	uint8_t a1, a2;
	//计算校验CRC
	_crc = CalcCrc(buf, (length - 2));
	//CRC低字节 
	a1 = _crc & 0xff;    
	//CRC高字节                 
	a2 = (_crc >> 8) & 0xff;               
	//校验正确
	if(a1 == buf[length - 2] && a2 == buf[length - 1]) 
	return 1;
	else
	return 0;
}
void timercallback(const ros::TimerEvent&)
{
    //单个磁导航模式
    if(Mode == 0)
    {
        //发送串口数据
        //01 03 00 04 00 16 85 C5
        SendData[0] = ID;SendData[1] = 0x03;
        SendData[2] = 0x00;SendData[3] = 0x04;
        SendData[4] = 0x00;SendData[5] = 0x16;
        uint16_t _crc = CalcCrc(SendData, 6);
        SendData[6] = _crc&0xff;SendData[7] = _crc>>8; 
    }
    else//多传感器集联模式
    {
         //发送串口数据
        //01 03 00 1F 00 1A CRC
        SendData[0] = ID;SendData[1] = 0x03;
        SendData[2] = 0x00;SendData[3] = 0x1F;
        SendData[4] = 0x00;SendData[5] = 0x1A;
        uint16_t _crc = CalcCrc(SendData, 6);
        SendData[6] = _crc&0xff;SendData[7] = _crc>>8;   
    }
    if(rev_ok)
    {
        ser.write(SendData,8);
        rev_ok = 0;
    }         
}
void DataAnalysis(uint8_t *buf)
{ 
    //单个磁导航模式
    if(Mode == 0)
    {
 
        Data.cdh.io = buf[3]<<8|buf[4];

        Data.cdh.offset_left = buf[37]<<8|buf[38];
        Data.cdh.offset_straight = buf[39]<<8|buf[40];
        Data.cdh.offset_right = buf[41]<<8|buf[42];
        Data.cdh.cdh_state = buf[43]<<8|buf[44];
        Data.cdh.cdh_allLedOn = buf[45]<<8|buf[46];        
    }
    else//多传感器集联模式
    {
        NetData.cdh11.io = buf[3]<<8|buf[4];
        NetData.cdh11.offset_left = buf[5]<<8|buf[6];
        NetData.cdh11.offset_straight = buf[7]<<8|buf[8];
        NetData.cdh11.offset_right = buf[9]<<8|buf[10];
        NetData.cdh11.cdh_state = buf[11]<<8|buf[12];
        NetData.cdh11.cdh_allLedOn = buf[13]<<8|buf[14];

        NetData.cdh12.io = buf[15]<<8|buf[16];
        NetData.cdh12.offset_left = buf[17]<<8|buf[18];
        NetData.cdh12.offset_straight = buf[19]<<8|buf[20];
        NetData.cdh12.offset_right = buf[21]<<8|buf[22];
        NetData.cdh12.cdh_state = buf[23]<<8|buf[24];
        NetData.cdh12.cdh_allLedOn = buf[25]<<8|buf[26];

        NetData.cdh13.io = buf[27]<<8|buf[28];
        NetData.cdh13.offset_left = buf[29]<<8|buf[30];
        NetData.cdh13.offset_straight = buf[31]<<8|buf[328];
        NetData.cdh13.offset_right = buf[33]<<8|buf[34];
        NetData.cdh13.cdh_state = buf[35]<<8|buf[36];
        NetData.cdh13.cdh_allLedOn = buf[37]<<8|buf[38];

        NetData.cdh14.io = buf[39]<<8|buf[40];
        NetData.cdh14.offset_left = buf[41]<<8|buf[42];
        NetData.cdh14.offset_straight = buf[43]<<8|buf[44];
        NetData.cdh14.offset_right = buf[45]<<8|buf[46];
        NetData.cdh14.cdh_state = buf[47]<<8|buf[48];
        NetData.cdh14.cdh_allLedOn = buf[49]<<8|buf[50];

        NetData.rfid5.code_h = buf[51]<<8|buf[52];
        NetData.rfid5.code_h = buf[53]<<8|buf[54];
    }
}
int main(int argc, char *argv[])
{
    setlocale(LC_ALL,"");
    ros::init(argc,argv,"igk_g416");
    ros::NodeHandle nh;
    ros::NodeHandle nh_private("~");
    //获取参数
    nh_private.param("Port",port,std::string("/dev/ttyUSB1")); 
    nh_private.param("RequestCycle",RequestCycle,0.02);
    nh_private.param("Baudrate",Baudrate,9600);
    nh_private.param("ID",ID,1);
    nh_private.param("Mode",Mode,0); 
    ros::Publisher pub;       
    if(Mode)
    {
        pub = nh.advertise<igk_g416::netdata>("igk_g416",10); 
    }
    else
    {
        pub = nh.advertise<igk_g416::data>("igk_g416",10);
    }

    ROS_INFO("端口号:%s,波特率:%d,请求周期:%.2f,模式:%d",port.c_str(),Baudrate,RequestCycle,Mode);
    ros::Timer timer1 = nh.createTimer(ros::Duration (RequestCycle), timercallback);    //设置定时器每隔20ms发送数据
    try 
    { 
        //设置串口属性，并打开串口 
        ser.setPort(port); 
        ser.setBaudrate(Baudrate); 
        serial::Timeout to = serial::Timeout::simpleTimeout(1000); 
        ser.setTimeout(to); 
        ser.open(); 
    } 
    catch (serial::IOException& e) 
    { 
        ROS_ERROR_STREAM("Unable to open port "); 
        return -1; 
    } 

    //检测串口是否已经打开，并给出提示信息 
    if(ser.isOpen()) 
    { 
        ROS_INFO_STREAM("Serial Port initialized"); 
    } 
    else 
    { 
        return -1; 
    } 
    ros::Rate rate(1/RequestCycle);
    uint8_t revBuf[100] ={0},timeout = 0; 

    while (ros::ok())       
    {
        timeout++;
        if(ser.available())
        { 
            size_t len = ser.available();
            ser.read(revBuf, len);
            if ((revBuf[0] == ID) && CheckModbusCRC(revBuf,len))
            {
                //数据解析
                DataAnalysis(revBuf);                    
                for (int i=0; i<len; i++)
                {
                    std::cout.fill('0');
                    std::cout.width(2);
                    std::cout << std::hex << (revBuf[i] & 0xff) << " ";
                }   
                std::cout << std::endl;
                Data.online = true;
                NetData.online = true;
                rev_ok = 1;
                timeout = 0;
            }
            else
            {                
                ROS_INFO("check error!");           
                rev_ok = 1;               
                size_t len = ser.available();
                //清空数据残余
                if(len > 0)
                {
                    ser.read(revBuf,len);
                }
            }
        }
        if(Mode)
        {
            NetData.header.stamp = ros::Time::now();
            NetData.header.frame_id = "IGK-Sensor";   
            pub.publish(NetData);
        }
        else
        {
            Data.header.stamp = ros::Time::now();
            Data.header.frame_id = "IGK-Sensor";  
            pub.publish(Data);
        }        
        if(timeout>(1/RequestCycle))
        {
            NetData.online = false;
            Data.online = false;
            timeout = 0;
            rev_ok = 1; 
        }
        rate.sleep();
        ros::spinOnce();
    }
    
    return 0;
}
